"""Spherical Gaussians"""
#  [BSD 2-CLAUSE LICENSE]
#
#  Copyright SVOX Authors 2021
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  1. Redistributions of source code must retain the above copyright notice,
#  this list of conditions and the following disclaimer.
#
#  2. Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.


import numpy as np
import math
import torch

def eval_sg(sg_lambda, sg_mu, sg_coeffs, dirs):
    """
    Evaluate spherical gaussians at unit directions
    using learnable SG basis.
    Works with torch.
    ... Can be 0 or more batch dimensions.
    N is the number of SG basis we use.
    :math:`Output = \sigma_{i}{exp ^ {\lambda_i * (\dot(\mu_i, \dirs) - 1)}`

    :param sg_lambda: The sharpness of the SG lobes. (N)
    :param sg_mu: The directions of the SG lobes. (N, 3)
    :param sg_coeffs: The coefficients of the SG (lob amplitude). (..., C, N)
    :param dirs: jnp.ndarray unit directions (..., 3)

    :return: (..., C)
    """
    
    product = torch.einsum(
        "ij,...j->...i", sg_mu, dirs)  # [..., N]
    basis = torch.exp(torch.einsum(
        "i,...i->...i", sg_lambda, product - 1))  # [..., N]
    output = torch.einsum(
        "...ki,...i->...k", sg_coeffs, basis)  # [..., C]
    output /= sg_mu.size(0)
    return output


def eval_sg_bases(sg_lambda, sg_mu, dirs):
    """
    Evaluate spherical gaussians bases at unit directions
    using learnable SG basis,
    without taking linear combination
    Works with torch.
    ... Can be 0 or more batch dimensions.
    N is the number of SG basis we use.
    :math:`Output = \sigma_{i}{exp ^ {\lambda_i * (\dot(\mu_i, \dirs) - 1)}`

    :param sg_lambda: The sharpness of the SG lobes. (N)
    :param sg_mu: The directions of the SG lobes. (N, 3)
    :param dirs: jnp.ndarray unit directions (..., 3)

    :return: (..., N)
    """
    product = torch.einsum(
        "ij,...j->...i", sg_mu, dirs)  # [..., N]
    basis = torch.exp(torch.einsum(
        "i,...i->...i", sg_lambda, product - 1))  # [..., N]
    return basis / sg_mu.size(0)



import svox
import torch

device = 'cuda:0'
t = svox.N3Tree(map_location=device, data_format="SH1")

t[0, 0, 0, :-1] = 0.0
t[0, 0, 0, -1:] = 0.5
r = svox.VolumeRenderer(t)

target =  torch.tensor([[0.0, 1.0, 0.5]], device=device)

ray_ori = torch.tensor([[0.1, 0.1, -0.1]], device=device)
ray_dir = torch.tensor([[0.0, 0.0, 1.0]], device=device)
ray = svox.Rays(origins=ray_ori, dirs=ray_dir, viewdirs=ray_dir)

lr = 2.5

print('GRADIENT DESC')

for i in range(10):
    rend = r(ray, cuda=True)
    if i % 2 == 0:
        print(rend.detach()[0].cpu().numpy())
    ((rend - target) ** 2).sum().backward()
    t.data.data -= lr * t.data.grad
    t.zero_grad()

t.expand("SH4")
print(r.data_format)
for i in range(10):
    rend = r(ray, cuda=True)
    if i % 2 == 0:
        print(rend.detach()[0].cpu().numpy())
    ((rend - target) ** 2).sum().backward()
    t.data.data -= lr * t.data.grad
    t.zero_grad()

print('TARGET')
print(target[0].cpu().numpy())



import svox
import torch
import torch.cuda
from matplotlib import pyplot as plt  

device = 'cuda:0'

t = svox.N3Tree.load("/home/sxyu/proj/volrend/build/lego_diffscale.npz",
                      map_location=device)
#  t = svox.N3Tree(map_location=device)
r = svox.VolumeRenderer(t)
c2w = torch.tensor([
                [ -0.9999999403953552, 0.0, 0.0, 0.0 ],
                [ 0.0, -0.7341099977493286, 0.6790305972099304, 2.737260103225708 ],
                [ 0.0, 0.6790306568145752, 0.7341098785400391, 2.959291696548462 ],
                [ 0.0, 0.0, 0.0, 1.0 ],
            ], device=device)

with torch.no_grad():
    start = torch.cuda.Event(enable_timing=True)
    end = torch.cuda.Event(enable_timing=True)
    im = None
    with t.accumulate_weights() as accum:
        start.record()
        for i in range(5):
            im = r.render_persp(c2w, height=800, width=800, fx=1111, cuda=True, fast=True)
        end.record()
    accum = accum()
    t[accum > 1, :1] = 100.0
    t[accum > 1, 1:16] = 0.0
    t[accum > 1, 16:3*16] = 0.0
    t[accum > 1, 3*16] = 100.0
    t.save('tmp.npz')

    torch.cuda.synchronize(device)
    dur = start.elapsed_time(end) / 5
    print('render time', dur, 'ms =', 1000 / dur, 'fps')
    print(im.shape)

    im = im.detach().clamp_(0.0, 1.0)
    plt.figure()
    plt.imshow(im.cpu())
    plt.show()
